//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/Data/DataUtil.cpp
//! @brief     Implements namespace DataUtil.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/Data/DataUtil.h"
#include "Base/Axis/Frame.h"
#include "Base/Axis/MakeScale.h"
#include "Base/Axis/Scale.h"
#include "Base/Util/Assert.h"
#include "Device/Data/Datafield.h"
#include <algorithm>
#include <cmath>
#include <functional>
#include <stdexcept>

std::vector<std::vector<double>>
DataUtil::invertAxis(int axis, const std::vector<std::vector<double>>& original)
{
    std::vector<std::vector<double>> inverse = original;

    size_t orig_rows = original.size();

    if (axis == 1) {
        for (size_t i = 0; i < orig_rows; ++i)
            inverse[i] = original[orig_rows - i - 1];
    } else if (axis == 0) {
        ASSERT(orig_rows > 0);
        size_t orig_cols = original.front().size();
        for (size_t i = 0; i < orig_rows; ++i)
            for (size_t j = 0; j < orig_cols; ++j)
                inverse[i][j] = original[i][orig_cols - j - 1];
    } else
        ASSERT_NEVER;

    return inverse;
}

std::vector<std::vector<double>>
DataUtil::transpose(const std::vector<std::vector<double>>& original)
{
    ASSERT(!original.empty());

    size_t orig_rows = original.size();
    size_t orig_cols = original.front().size();

    std::vector<std::vector<double>> transposed(orig_cols, std::vector<double>(orig_rows));

    for (size_t i = 0; i < orig_rows; ++i)
        for (size_t j = 0; j < orig_cols; ++j)
            transposed[j][i] = original[i][j];

    return transposed;
}

Datafield DataUtil::createRearrangedDataSet(const Datafield& data, int n)
{
    ASSERT(data.rank() == 2);
    n = (4 + n % 4) % 4;
    if (n == 0)
        return data;

    std::unique_ptr<Datafield> output;
    std::function<void(std::vector<int>&)> index_mapping;

    if (n == 2) {
        output.reset(new Datafield({data.axis(0).clone(), data.axis(1).clone()}));
        const int end_bin_x = static_cast<int>(data.axis(0).size()) - 1;
        const int end_bin_y = static_cast<int>(data.axis(1).size()) - 1;
        index_mapping = [end_bin_x, end_bin_y](std::vector<int>& inds) {
            inds[0] = end_bin_x - inds[0];
            inds[1] = end_bin_y - inds[1];
        };

    } else {
        output.reset(new Datafield({data.axis(1).clone(), data.axis(0).clone()}));
        const size_t rev_axis_i = n % 3;
        const size_t end_bin = data.axis(rev_axis_i).size() - 1;
        index_mapping = [rev_axis_i, end_bin](std::vector<int>& inds) {
            const int tm_index = inds[rev_axis_i];
            inds[rev_axis_i] = inds[rev_axis_i ^ 1];
            inds[rev_axis_i ^ 1] = static_cast<int>(end_bin) - tm_index;
        };
    }

    for (size_t i = 0, size = data.size(); i < size; ++i) {
        std::vector<int> axis_inds = data.frame().allIndices(i);
        index_mapping(axis_inds);
        size_t iout = output->frame().toGlobalIndex(
            {static_cast<unsigned>(axis_inds[0]), static_cast<unsigned>(axis_inds[1])});
        (*output)[iout] = data[i];
    }
    return *output;
}

Datafield DataUtil::vecvecToDatafield(const std::vector<std::vector<double>>& array_2d)
{
    size_t nrows = array_2d.size();
    size_t ncols = array_2d[0].size();

    std::vector<const Scale*> axes{newEquiDivision("u! (bin)", nrows, 0.0, double(nrows)),
                                   newEquiDivision("v! (bin)", ncols, 0.0, double(ncols))};
    std::vector<double> out;
    out.reserve(nrows * ncols);
    for (size_t row = 0; row < nrows; row++) {
        for (size_t col = 0; col < ncols; col++)
            out.push_back(array_2d[row][col]);
    }
    return {std::move(axes), out};
}
