//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Device/IO/ReadWrite2DTable.cpp
//! @brief     Implements functions read|writeNumpyTxt
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Device/IO/ReadWrite2DTable.h"
#include "Base/Axis/MakeScale.h"
#include "Base/Axis/Scale.h"
#include "Base/Util/Assert.h"
#include "Base/Util/StringUtil.h"
#include "Device/Data/DataUtil.h"
#include "Device/Data/Datafield.h"
#include "Device/IO/ParseUtil.h"
#include <string>
#include <vector>

namespace {

bool isDoubleStartChar(char c)
{
    return isdigit(c) || c == '-' || c == '+';
}

void write1DRepresentation(const Datafield& data, std::ostream& output_stream)
{
    output_stream << "# coordinates         intensities" << std::endl;
    output_stream.imbue(std::locale::classic());
    output_stream << std::scientific << std::setprecision(12);

    const std::vector<double> axis_values = data.axis(0).binCenters();

    // printing coordinate and associated intensity
    for (size_t i = 0, nrows = axis_values.size(); i < nrows; ++i)
        output_stream << axis_values[i] << "    " << Util::Parse::ignoreDenormalized(data[i])
                      << std::endl;
}

void write2DRepresentation(const Datafield& data, std::ostream& output_stream)
{
    const size_t nrows = data.axis(1).size();
    const size_t ncols = data.axis(0).size();

    output_stream << "# [nrows=" << nrows << ", ncols=" << ncols << "]" << std::endl;

    std::vector<std::vector<double>> dataArray = data.values2D();
    output_stream.imbue(std::locale::classic());
    output_stream << std::scientific << std::setprecision(12);

    for (size_t i = 0; i < nrows; i++) {
        for (size_t j = 0; j < ncols; j++) {
            double z_value = dataArray[i][j];
            output_stream << Util::Parse::ignoreDenormalized(z_value) << "    ";
        }
        output_stream << std::endl;
    }
}

} // namespace

Datafield* Util::RW::read2DTable(std::istream& input_stream)
{
    std::string line;
    std::vector<double> values;

    // Read numbers from input stream:
    size_t nrows = 0;
    size_t ncols = 0;
    while (std::getline(input_stream, line)) {
        line = Base::String::trim(line);
        if (line.empty() || !isDoubleStartChar(line[0]))
            continue;
        std::vector<double> tmp = Util::Parse::parse_doubles(line);
        if (nrows == 0)
            ncols = tmp.size();
        else if (tmp.size() != ncols)
            throw std::runtime_error("Number of elements is not the same for all rows");
        for (const double val : tmp)
            values.push_back(val);
        ++nrows;
    }
    if (nrows == 0 || ncols == 0)
        throw std::runtime_error("No data found in table");

    std::vector<const Scale*> axes;

    if (nrows < 2)
        axes = {newEquiDivision("axis0", ncols, 0.0, (double)ncols)};
    else if (ncols < 2)
        axes = {newEquiDivision("axis0", nrows, 0.0, (double)nrows)};
    else
        axes = {newEquiDivision("axis0", ncols, 0.0, (double)ncols),
                newEquiDivision("axis1", nrows, 0.0, (double)nrows)};

    return new Datafield(std::move(axes), values);
}

void Util::RW::write2DTable(const Datafield& data, std::ostream& output_stream)
{
    output_stream << "# BornAgain Intensity Data" << std::endl;
    output_stream << "# Simple array suitable for numpy, matlab etc." << std::endl;

    const size_t dim = data.rank();
    switch (dim) {
    case 1:
        write1DRepresentation(data, output_stream);
        break;
    case 2:
        write2DRepresentation(data, output_stream);
        break;
    default:
        ASSERT_NEVER;
    }
}
